@@ -22,7 +22,7 @@ along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 """
 
-__version__ = "0.3.40"
+__version__ = "0.3.42"
 __author__ = "nagev"
 __license__ = "GPLv3"
 
@@ -57,18 +57,6 @@ if os.environ.get("pafydebug") == "1":
 dbg = logging.debug
 
 
-def _make_url(raw, sig, quick=True):
-    """ Return usable url. Set quick=False to disable ratebypass override. """
-
-    if quick and not "ratebypass=" in raw:
-        raw += "&ratebypass=yes"
-
-    if not "signature=" in raw:
-        raw += "&signature=" + sig
-
-    return raw
-
-
 def new(url, basic=True, gdata=False, signature=True, size=False,
         callback=None):
     """ Return a new pafy instance given a url or video id.
@@ -103,6 +91,49 @@ def new(url, basic=True, gdata=False, signature=True, size=False,
     return Pafy(url, basic, gdata, signature, size, callback)
 
 
+def get_video_info(video_id, newurl=None):
+    """ Return info for video_id.  Returns dict. """
+
+    url = g.urls['vidinfo'] % video_id
+    url = newurl if newurl else url
+    info = decode_if_py3(g.opener.open(url).read())
+    info = parse_qs(info)
+    dbg("Fetched video info")
+
+    if info['status'][0] == "fail":
+        reason = info['reason'][0] or "Bad video argument"
+        raise IOError("Youtube says: %s [%s]" % (reason, video_id))
+
+    return info
+
+
+def get_video_gdata(video_id):
+    """ Return xml string containing video metadata from gdata api. """
+
+    new.callback("Fetching video gdata")
+    url = g.urls['gdata'] % video_id
+    gdata = g.opener.open(url).read()
+    dbg("Fetched video gdata")
+    new.callback("Fetched video gdata")
+    return gdata
+
+
+def extract_video_id(url):
+    """ Extract the video id from a url, return video id as str. """
+
+    ok = (r"\w-",) * 3
+    regx = re.compile(r'(?:^|[^%s]+)([%s]{11})(?:[^%s]+|$)' % ok)
+    url = str(url)
+    m = regx.search(url)
+
+    if not m:
+        err = "Need 11 character video id or the URL of the video. Got %s"
+        raise ValueError(err % url)
+
+    vidid = m.group(1)
+    return vidid
+
+
 class g(object):
 
     """ Class for holding constants needed throughout the module. """
@@ -113,13 +144,16 @@ class g(object):
         'vidinfo': ('http://www.youtube.com/get_video_info?'
                     'video_id=%s&asv=3&el=detailpage&hl=en_US'),
         'playlist': ('http://www.youtube.com/list_ajax?'
-                     'style=json&action_get_list=1&list=%s')
+                     'style=json&action_get_list=1&list=%s'),
+        'age_vidinfo': ('https://www.youtube.com/get_video_info?video_id=%s&'
+                        'el=player_embedded&gl=US&hl=en&eurl=https://youtube.'
+                        'googleapis.com/v/%s&asv=3&sts=1588')
     }
     ua = ("Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64;"
           " Trident/5.0)")
     UEFSM = 'url_encoded_fmt_stream_map'
     AF = 'adaptive_fmts'
-    jsplayer = r';ytplayer.config = ({.*?});'
+    jsplayer = r';\s*ytplayer\.config\s*=\s*(\s*{.*?}\s*)\s*;'
     lifespan = 60 * 60 * 5  # 5 hours
     opener = build_opener()
     opener.addheaders = [('User-Agent', ua)]
@@ -137,6 +171,10 @@ class g(object):
         '44': ('854x480', 'webm', "normal", ''),
         '45': ('1280x720', 'webm', "normal", ''),
         '46': ('1920x1080', 'webm', "normal", ''),
+
+        #'59': ('1x1', 'mp4', 'normal', ''),
+        #'78': ('1x1', 'mp4', 'normal', ''),
+
         '82': ('640x360-3D', 'mp4', "normal", ''),
         '83': ('640x480-3D', 'mp4', 'normal', ''),
         '84': ('1280x720-3D', 'mp4', "normal", ''),
@@ -158,16 +196,16 @@ class g(object):
         '170': ('1920x1080', 'webm', 'video', ''),
         '171': ('128k', 'ogg', 'audio', ''),
         '172': ('192k', 'ogg', 'audio', ''),
-        '242': ('360x240', 'webm', 'normal', ''),
-        '243': ('480x360', 'webm', 'normal', ''),
-        '244': ('640x480', 'webm', 'normal', ''),
-        '245': ('640x480', 'webm', 'normal', ''),
-        '246': ('640x480', 'webm', 'normal', ''),
-        '247': ('720x480', 'webm', 'normal', ''),
-        '248': ('1920x1080', 'webm', 'normal', ''),
+        '242': ('360x240', 'webm', 'video', 'VP9'),
+        '243': ('480x360', 'webm', 'video', 'VP9'),
+        '244': ('640x480', 'webm', 'video', 'VP9'),
+        '245': ('640x480', 'webm', 'video', 'VP9'),
+        '246': ('640x480', 'webm', 'video', 'VP9'),
+        '247': ('720x480', 'webm', 'video', 'VP9'),
+        '248': ('1920x1080', 'webm', 'video', 'VP9'),
         '256': ('192k', 'm4a', 'audio', '6-channel'),
         '258': ('320k', 'm4a', 'audio', '6-channel'),
-        '264': ('1920x1080', 'm4v', 'video', '')
+        '264': ('2560x1440', 'm4v', 'video', '')
     }
 
 
@@ -180,18 +218,15 @@ def _extract_smap(map_name, dic, zero_idx=True):
         smap = smap.split(",")
         smap = [parse_qs(x) for x in smap]
 
-        if not early_py_version:
-            return [{k: v[0] for k, v in x.items()} for x in smap]
-
-        else:
-            # Python 2.6 compatibility
-            r = []
-            for x in smap:
-                z = {}
-                for k, v in x.items():
-                    z[k] = v[0]
-                r.append(z)
-            return r
+        # Python 2.6 compatibility
+        #return [{k: v[0] for k, v in x.items()} for x in smap]
+        r = []
+        for x in smap:
+            z = {}
+            for k, v in x.items():
+                z[k] = v[0]
+            r.append(z)
+        return r
 
     return []
 
@@ -373,48 +408,6 @@ def _decodesig(sig, js_url):
     return solved
 
 
-def extract_video_id(url):
-    """ Extract the video id from a url, return video id as str. """
-
-    ok = (r"\w-",) * 3
-    regx = re.compile(r'(?:^|[^%s]+)([%s]{11})(?:[^%s]+|$)' % ok)
-    url = str(url)
-    m = regx.search(url)
-
-    if not m:
-        err = "Need 11 character video id or the URL of the video. Got %s"
-        raise ValueError(err % url)
-
-    vidid = m.group(1)
-    return vidid
-
-
-def get_video_info(video_id):
-    """ Return info for video_id.  Returns dict. """
-
-    url = g.urls['vidinfo'] % video_id
-    info = decode_if_py3(g.opener.open(url).read())
-    info = parse_qs(info)
-    dbg("Fetched video info")
-
-    if info['status'][0] == "fail":
-        reason = info['reason'][0] or "Bad video argument"
-        raise IOError("Youtube says: %s [%s]" % (reason, video_id))
-
-    return info
-
-
-def get_video_gdata(video_id):
-    """ Return xml string containing video metadata from gdata api. """
-
-    new.callback("Fetching video gdata")
-    url = g.urls['gdata'] % video_id
-    gdata = g.opener.open(url).read()
-    dbg("Fetched video gdata")
-    new.callback("Fetched video gdata")
-    return gdata
-
-
 def get_js_sm(video_id):
     """ Fetch watchinfo page and extract stream map and js funcs if not known.
 
@@ -429,6 +422,15 @@ def get_js_sm(video_id):
     watch_url = g.urls['watchv'] % video_id
     new.callback("Fetching watchv page")
     watchinfo = g.opener.open(watch_url).read().decode("utf8")
+
+    if re.search(r'player-age-gate-content">', watchinfo) is not None:
+        # create a new Pafy object
+        dbg("creating new instance for age restrictved video")
+        doppleganger = new(video_id, False, False, False)
+        video_info_url = g.urls['age_vidinfo'] % (video_id, video_id)
+        doppleganger._fetch_basic(ageurl=video_info_url)
+        return "age", "age", doppleganger
+
     dbg("Fetched watchv page")
     new.callback("Fetched watchv page")
     m = re.search(g.jsplayer, watchinfo)
@@ -452,6 +454,30 @@ def get_js_sm(video_id):
     return smap, js_url, funcs
 
 
+def _make_url(raw, sig, quick=True):
+    """ Return usable url. Set quick=False to disable ratebypass override. """
+
+    if quick and not "ratebypass=" in raw:
+        raw += "&ratebypass=yes"
+
+    if not "signature=" in raw:
+        raw += "&signature=" + sig
+
+    return raw
+
+
+def gen_ageurl(dop, itag):
+    """ Decrypt signature for age-restricted item. Return url """
+
+    for x in dop._sm + dop._asm:
+
+        if x['itag'] == itag and len(x['s']) == 86:
+            s = x['s']
+            s = s[2:63] + s[82] + s[64:82] + s[63]
+            dbg("decrypted agesig: %s%s", s[:22], "..")
+            return _make_url(x['url'], s)
+
+
 def _get_matching_stream(smap, itag):
     """ Return the url and signature for a stream matching itag in smap. """
 
@@ -589,12 +615,14 @@ class Stream(object):
         if self._url:
             pass
 
+        elif self._parent.age:
+            self._url = gen_ageurl(self._parent.doppleganger, self.itag)
+
         elif not self.encrypted:
             self._url = _make_url(self._rawurl, self._sig)
 
         else:
             # encrypted url signatures
-
             if self._parent.js_url:
                 #dbg("using cached js %s" % self._parent.js_url[-15:])
                 enc_streams = self._parent.enc_streams
@@ -604,6 +632,14 @@ class Stream(object):
                 self._parent.expiry = time.time() + g.lifespan
                 self._parent.js_url = js_url
 
+                # check for age
+                if type(enc_streams) == str and enc_streams == "age":
+                    self._parent.age = True
+                    dop = self._parent.doppleganger = funcs
+                    self._url = gen_ageurl(dop, self.itag)
+                    return self._url
+
+
                 # Create Pafy funcmap dict for this js_url
                 if not Pafy.funcmap.get(js_url):
                     Pafy.funcmap[js_url] = funcs
@@ -653,7 +689,7 @@ class Stream(object):
 
         if early_py_version:
             status_string = ('  {0:} Bytes [{1:.2%}] received. Rate:'
-                                 ' [{2:4.0f} KB/s].  ETA: [{3:.0f} secs]')
+                             ' [{2:4.0f} KB/s].  ETA: [{3:.0f} secs]')
 
         response = g.opener.open(self.url)
         total = int(response.info()['Content-Length'].strip())
@@ -666,7 +702,7 @@ class Stream(object):
         except IOError:
             ok = re.compile(r'[^\\/?*$\'"%&:<>|]')
             fname = "".join(x if ok.match(x) else "_" for x in self.filename)
-            outfh = open(fname, 'wb')
+            outfh = open(fname.encode("utf8", errors="ignore"), 'wb')
 
         while True:
             chunk = response.read(chunksize)
@@ -722,6 +758,7 @@ class Pafy(object):
         self._sm = []
         self._asm = []
         self.js_url = None  # if js_url is set then has new stream map
+        self.age = False
         self._streams = []
         self._oggstreams = []
         self._m4astreams = []
@@ -764,20 +801,26 @@ class Pafy(object):
                 # pylint: disable=W0104
                 s.get_filesize()
 
-    def _fetch_basic(self):
+    def _fetch_basic(self, ageurl=None):
         """ Fetch info url page and set member vars. """
 
         if self._have_basic:
             return
 
-        allinfo = get_video_info(self.videoid)
+        if ageurl:
+            allinfo = get_video_info("none", ageurl)
+
+        else:
+            allinfo = get_video_info(self.videoid)
+
         new.callback("Fetched video info")
         f = lambda x: allinfo.get(x, ["unknown"])[0]
+        t = lambda x: allinfo.get(x, [0.0])[0]
         z = lambda x: allinfo.get(x, [""])[0]
         self._title = f('title').replace("/", "-")
         self._author = f('author')
         self._videoid = f('video_id')
-        self._rating = float(f('avg_rating'))
+        self._rating = float(t('avg_rating'))
         self._length = int(f('length_seconds'))
         self._viewcount = int(f('view_count'))
         self._thumb = unquote_plus(f('thumbnail_url'))
@@ -787,6 +830,10 @@ class Pafy(object):
         self._bigthumbhd = z('iurlsdmaxres')
         self.ciphertag = f("use_cipher_signature") == "True"
 
+        if ageurl:
+            self.ciphertag = False
+            dbg("Encrypted signature detected - age restricted")
+
         if self.ciphertag:
             dbg("Encrypted signature detected.")
 
@@ -807,9 +854,13 @@ class Pafy(object):
         gdata = get_video_gdata(self.videoid)
         t0 = "{http://search.yahoo.com/mrss/}"
         t1 = "{http://www.w3.org/2005/Atom}"
+        t2 = "{http://gdata.youtube.com/schemas/2007}"
         tree = ElementTree.fromstring(gdata)
         groups = tree.find(t0 + "group")
         published = tree.find(t1 + "published").text
+        rating = tree.find(t2 + "rating")
+        likes = rating.get("numLikes") if rating is not None else 0
+        dislikes = rating.get("numDislikes") if rating is not None else 0
         description = groups.find(t0 + "description").text
         category = groups.find(t0 + "category").text
         username = tree.find(t1 + "author/" + t1 + "uri").text.split("/")[-1]
@@ -817,6 +868,8 @@ class Pafy(object):
         setattr(self, "_published", published)
         setattr(self, "_description", description)
         setattr(self, "_category", category)
+        setattr(self, "_likes", likes)
+        setattr(self, "_dislikes", dislikes)
 
         self._have_gdata = 1
 
@@ -1012,7 +1065,21 @@ class Pafy(object):
         """ The upload date and time of the video. Returns string. """
 
         self._fetch_gdata()
-        return self._published.rstrip(".000Z").replace("T", " ")
+        return self._published.replace(".000Z", "").replace("T", " ")
+
+    @property
+    def likes(self):
+        """ The number of likes for the video. Returns int. """
+
+        self._fetch_gdata()
+        return int(self._likes)
+
+    @property
+    def dislikes(self):
+        """ The number of dislikes for the video. Returns int. """
+
+        self._fetch_gdata()
+        return int(self._dislikes)
 
     def getbest(self, preftype="any", ftypestrict=True):
         """
@@ -1073,7 +1140,7 @@ class Pafy(object):
         self._author = pl_data.get("author")
         self._length = int(pl_data.get("length_seconds", 0))
         self._rating = pl_data.get("rating", 0.0)
-        self._viewcount = "".join(re.findall("\d", pl_data.get("views", "0")))
+        self._viewcount = "".join(re.findall(r"\d", pl_data.get("views", "0")))
         self._viewcount = int(self._viewcount)
         self._thumb = pl_data.get("thumbnail")
         self._description = pl_data.get("description")
@@ -1110,6 +1177,7 @@ def get_playlist(playlist_url, basic=False, gdata=False, signature=False,
     except:
         raise IOError("Error fetching playlist %s" % m.groups(0))
 
+    # playlist specific metadata
     playlist = dict(
         playlist_id=playlist_id,
         likes=allinfo.get('likes'),
@@ -1120,6 +1188,7 @@ def get_playlist(playlist_url, basic=False, gdata=False, signature=False,
         items=[]
     )
 
+    # playlist items specific metadata
     for v in allinfo['video']:
 
         vid_data = dict(
